@@ -2,7 +2,8 @@
   +----------------------------------------------------------------------+
   | Suhosin Version 1                                                    |
   +----------------------------------------------------------------------+
-  | Copyright (c) 2006 The Hardened-PHP Project                          |
+  | Copyright (c) 2006-2007 The Hardened-PHP Project                     |
+  | Copyright (c) 2007-2014 SektionEins GmbH                             |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
@@ -12,26 +13,35 @@
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
-  | Author: Stefan Esser <sesser@hardened-php.net>                       |
+  | Author: Stefan Esser <sesser@sektioneins.de>                         |
   +----------------------------------------------------------------------+
 */
 
-/* $Id: execute.c,v 1.43 2007-04-12 09:37:43 sesser Exp $ */
+/* $Id: execute.c,v 1.1.1.1 2007-11-28 01:15:35 sesser Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
 
+#include <fcntl.h>
 #include "php.h"
 #include "php_ini.h"
 #include "zend_hash.h"
 #include "zend_extensions.h"
 #include "ext/standard/info.h"
+#include "ext/standard/php_rand.h"
+#include "ext/standard/php_lcg.h"
 #include "php_suhosin.h"
 #include "zend_compile.h"
 #include "zend_llist.h"
 #include "SAPI.h"
 
+#include "sha256.h"
+
+#if PHP_VERSION_ID >= 50500
+static void (*old_execute_ex)(zend_execute_data *execute_data TSRMLS_DC);
+static void suhosin_execute_ex(zend_execute_data *execute_data TSRMLS_DC);
+#endif
 
 static void (*old_execute)(zend_op_array *op_array TSRMLS_DC);
 static void suhosin_execute(zend_op_array *op_array TSRMLS_DC);
@@ -42,8 +52,13 @@ static void *(*zo_set_oe_ex)(void *ptr) = NULL;
 /*STATIC zend_op_array* (*old_compile_file)(zend_file_handle* file_handle, int type TSRMLS_DC);
   STATIC zend_op_array* suhosin_compile_file(zend_file_handle*, int TSRMLS_DC);*/
 
+#if PHP_VERSION_ID >= 50500
+static void suhosin_execute_internal(zend_execute_data *execute_data_ptr, zend_fcall_info *fci, int return_value_used TSRMLS_DC);
+static void (*old_execute_internal)(zend_execute_data *execute_data_ptr, zend_fcall_info *fci, int return_value_used TSRMLS_DC);
+#else
 static void suhosin_execute_internal(zend_execute_data *execute_data_ptr, int return_value_used TSRMLS_DC);
 static void (*old_execute_internal)(zend_execute_data *execute_data_ptr, int return_value_used TSRMLS_DC);
+#endif
 
 extern zend_extension suhosin_zend_extension_entry;
 
@@ -83,6 +98,8 @@ conts:
 #define SUHOSIN_CODE_TYPE_BADFILE	12
 #define SUHOSIN_CODE_TYPE_LONGNAME	13
 #define SUHOSIN_CODE_TYPE_MANYDOTS	14
+#define SUHOSIN_CODE_TYPE_WRITABLE      15
+#define SUHOSIN_CODE_TYPE_MBREGEXP	16
 
 static int suhosin_check_filename(char *s, int len TSRMLS_DC)
 {
@@ -113,7 +130,7 @@ static int suhosin_check_filename(char *s, int len TSRMLS_DC)
 			return SUHOSIN_CODE_TYPE_UPLOADED;
 		}
 	}
-	
+		
 	/* count number of directory traversals */
 	for (i=0; i < len-3; i++) {
 		if (s[i] == '.' && s[i+1] == '.' && (s[i+2] == '/' || s[i+2] == '\\')) {
@@ -145,7 +162,7 @@ SDEBUG("xxx %08x %08x",SUHOSIN_G(include_whitelist),SUHOSIN_G(include_blacklist)
 			t = h = (h == NULL) ? h2 : ( (h2 == NULL) ? h : ( (h < h2) ? h : h2 ) );
 			if (h == NULL) break;
 							
-			while (t > s && (isalnum(t[-1]) || t[-1]=='_')) {
+			while (t > s && (isalnum(t[-1]) || t[-1]=='_' || t[-1]=='.')) {
 				t--;
 			}
 			
@@ -188,7 +205,7 @@ SDEBUG("xxx %08x %08x",SUHOSIN_G(include_whitelist),SUHOSIN_G(include_blacklist)
 			t = h = (h == NULL) ? h2 : ( (h2 == NULL) ? h : ( (h < h2) ? h : h2 ) );
 			if (h == NULL) break;
 							
-			while (t > s && (isalnum(t[-1]) || t[-1]=='_')) {
+			while (t > s && (isalnum(t[-1]) || t[-1]=='_' || t[-1]=='.')) {
 				t--;
 			}
 
@@ -215,6 +232,16 @@ SDEBUG("xxx %08x %08x",SUHOSIN_G(include_whitelist),SUHOSIN_G(include_blacklist)
 			s = h + 1;
 		} while (1);
 	}
+
+	/* disallow writable files */
+	if (!SUHOSIN_G(executor_include_allow_writable_files)) {
+	        /* protection against *REMOTE* attacks, potential
+	           race condition of access() is irrelevant */
+	        if (access(s, W_OK) == 0) {
+                        return SUHOSIN_CODE_TYPE_WRITABLE;
+	        }
+	}
+
 	return SUHOSIN_CODE_TYPE_GOODFILE;
 }
 
@@ -254,6 +281,11 @@ static zend_bool suhosin_zend_open(const char *filename, zend_file_handle *fh)
 			suhosin_log(S_INCLUDE, "Include filename contains an ASCIIZ character");
 			suhosin_bailout(TSRMLS_C);
 			break;
+		
+		    case SUHOSIN_CODE_TYPE_WRITABLE:
+			suhosin_log(S_INCLUDE, "Include filename ('%s') is writable by PHP process", filename);
+			suhosin_bailout(TSRMLS_C);
+			break;		    	
 
 		    case SUHOSIN_CODE_TYPE_BLACKURL:
 			suhosin_log(S_INCLUDE, "Include filename ('%s') is an URL that is forbidden by the blacklist", filename);
@@ -284,7 +316,7 @@ static int suhosin_detect_codetype(zend_op_array *op_array TSRMLS_DC)
 	char *s;
 	int r;
 
-	s = op_array->filename;
+	s = (char *)op_array->filename;
 	
 	/* eval, assert, create_function, preg_replace  */
 	if (op_array->type == ZEND_EVAL_CODE) {
@@ -301,6 +333,10 @@ static int suhosin_detect_codetype(zend_op_array *op_array TSRMLS_DC)
 			return SUHOSIN_CODE_TYPE_REGEXP;
 		}
 
+		if (strstr(s, "mbregex replace") != NULL) {
+			return SUHOSIN_CODE_TYPE_MBREGEXP;
+		}
+
 		if (strstr(s, "assert code") != NULL) {
 			return SUHOSIN_CODE_TYPE_ASSERT;
 		}
@@ -312,6 +348,18 @@ static int suhosin_detect_codetype(zend_op_array *op_array TSRMLS_DC)
 		if (strstr(s, "Command line code") != NULL) {
 			return SUHOSIN_CODE_TYPE_COMMANDLINE;
 		}
+
+		if (strstr(s, "Command line begin code") != NULL) {
+			return SUHOSIN_CODE_TYPE_COMMANDLINE;
+		}
+
+		if (strstr(s, "Command line run code") != NULL) {
+			return SUHOSIN_CODE_TYPE_COMMANDLINE;
+		}
+
+		if (strstr(s, "Command line end code") != NULL) {
+			return SUHOSIN_CODE_TYPE_COMMANDLINE;
+		}
 		
 		if (strstr(s, "suhosin internal code") != NULL) {
 			return SUHOSIN_CODE_TYPE_SUHOSIN;
@@ -333,8 +381,14 @@ static int suhosin_detect_codetype(zend_op_array *op_array TSRMLS_DC)
 
 /* {{{ void suhosin_execute_ex(zend_op_array *op_array TSRMLS_DC)
  *    This function provides a hook for execution */
+#if PHP_VERSION_ID > 50500
+static void suhosin_execute_ex(zend_execute_data *execute_data TSRMLS_DC)
+{
+	zend_op_array *op_array = execute_data->op_array;
+#else
 static void suhosin_execute_ex(zend_op_array *op_array, int zo, long dummy TSRMLS_DC)
 {
+#endif
 	zend_op_array *new_op_array;
 	int op_array_type, len;
 	char *fn;
@@ -342,6 +396,15 @@ static void suhosin_execute_ex(zend_op_array *op_array, int zo, long dummy TSRML
 	zend_uint orig_code_type;
 	unsigned long *suhosin_flags = NULL;
 	
+	/* log variable dropping statistics */
+	if (SUHOSIN_G(abort_request) && (SUHOSIN_G(att_request_variables)-SUHOSIN_G(cur_request_variables) > 0)) {
+		suhosin_log(S_VARS, "dropped %u request variables - (%u in GET, %u in POST, %u in COOKIE)",
+		SUHOSIN_G(att_request_variables)-SUHOSIN_G(cur_request_variables),
+		SUHOSIN_G(att_get_vars)-SUHOSIN_G(cur_get_vars),
+		SUHOSIN_G(att_post_vars)-SUHOSIN_G(cur_post_vars),
+		SUHOSIN_G(att_cookie_vars)-SUHOSIN_G(cur_cookie_vars));
+	}
+	
 	if (SUHOSIN_G(abort_request) && !SUHOSIN_G(simulation) && SUHOSIN_G(filter_action)) {
 	
 		char *action = SUHOSIN_G(filter_action);
@@ -431,11 +494,11 @@ static void suhosin_execute_ex(zend_op_array *op_array, int zo, long dummy TSRML
 	SUHOSIN_G(execution_depth)++;
 	
 	if (SUHOSIN_G(max_execution_depth) && SUHOSIN_G(execution_depth) > SUHOSIN_G(max_execution_depth)) {
-		suhosin_log(S_EXECUTOR, "maximum execution depth reached - script terminated");
+		suhosin_log(S_EXECUTOR|S_GETCALLER, "maximum execution depth reached - script terminated");
 		suhosin_bailout(TSRMLS_C);
 	}
 	
-	fn = op_array->filename;
+	fn = (char *)op_array->filename;
 	len = strlen(fn);
 	
 	orig_code_type = SUHOSIN_G(in_code_type);
@@ -478,7 +541,7 @@ not_evaled_code:
 	switch (op_array_type) {
 	    case SUHOSIN_CODE_TYPE_EVAL:
 		    if (SUHOSIN_G(executor_disable_eval)) {
-			    suhosin_log(S_EXECUTOR, "use of eval is forbidden by configuration");
+			    suhosin_log(S_EXECUTOR|S_GETCALLER, "use of eval is forbidden by configuration");
 			    if (!SUHOSIN_G(simulation)) {
 				    zend_error(E_ERROR, "SUHOSIN - Use of eval is forbidden by configuration");
 			    }
@@ -487,13 +550,17 @@ not_evaled_code:
 		    
 	    case SUHOSIN_CODE_TYPE_REGEXP:
 		    if (SUHOSIN_G(executor_disable_emod)) {
-			    suhosin_log(S_EXECUTOR, "use of preg_replace() with /e modifier is forbidden by configuration");
+			    suhosin_log(S_EXECUTOR|S_GETCALLER, "use of preg_replace() with /e modifier is forbidden by configuration");
 			    if (!SUHOSIN_G(simulation)) {
 				    zend_error(E_ERROR, "SUHOSIN - Use of preg_replace() with /e modifier is forbidden by configuration");
 			    }
 		    }
 		    break;
 		    
+		case SUHOSIN_CODE_TYPE_MBREGEXP:
+			/* XXX TODO: Do we want to disallow this, too? */
+			break;
+		
 	    case SUHOSIN_CODE_TYPE_ASSERT:
 		    break;
 		    
@@ -501,32 +568,37 @@ not_evaled_code:
 		    break;
 		    
 		case SUHOSIN_CODE_TYPE_LONGNAME:
-			suhosin_log(S_INCLUDE, "Include filename ('%s') is too long", op_array->filename);
+			suhosin_log(S_INCLUDE|S_GETCALLER, "Include filename ('%s') is too long", op_array->filename);
 			suhosin_bailout(TSRMLS_C);
 			break;
 
 		case SUHOSIN_CODE_TYPE_MANYDOTS:
-			suhosin_log(S_INCLUDE, "Include filename ('%s') contains too many '../'", op_array->filename);
+			suhosin_log(S_INCLUDE|S_GETCALLER, "Include filename ('%s') contains too many '../'", op_array->filename);
 			suhosin_bailout(TSRMLS_C);
 			break;
 	    
 		case SUHOSIN_CODE_TYPE_UPLOADED:
-		    suhosin_log(S_INCLUDE, "Include filename is an uploaded file");
+		    suhosin_log(S_INCLUDE|S_GETCALLER, "Include filename is an uploaded file");
 		    suhosin_bailout(TSRMLS_C);
 		    break;
 		    
 	    case SUHOSIN_CODE_TYPE_0FILE:
-			suhosin_log(S_INCLUDE, "Include filename contains an ASCIIZ character");
+			suhosin_log(S_INCLUDE|S_GETCALLER, "Include filename contains an ASCIIZ character");
 			suhosin_bailout(TSRMLS_C);
 			break;
+			
+    		case SUHOSIN_CODE_TYPE_WRITABLE:
+    		        suhosin_log(S_INCLUDE|S_GETCALLER, "Include filename ('%s') is writable by PHP process", op_array->filename);
+    			suhosin_bailout(TSRMLS_C);
+    			break;		    	
 
 	    case SUHOSIN_CODE_TYPE_BLACKURL:
-			suhosin_log(S_INCLUDE, "Include filename ('%s') is an URL that is forbidden by the blacklist", op_array->filename);
+			suhosin_log(S_INCLUDE|S_GETCALLER, "Include filename ('%s') is an URL that is forbidden by the blacklist", op_array->filename);
 			suhosin_bailout(TSRMLS_C);
 			break;
 			
 	    case SUHOSIN_CODE_TYPE_BADURL:
-			suhosin_log(S_INCLUDE, "Include filename ('%s') is an URL that is not allowed", op_array->filename);
+			suhosin_log(S_INCLUDE|S_GETCALLER, "Include filename ('%s') is an URL that is not allowed", op_array->filename);
 		    suhosin_bailout(TSRMLS_C);
 			break;
 
@@ -552,17 +624,22 @@ not_evaled_code:
 	}
 
 continue_execution:
+#if PHP_VERSION_ID >= 50500
+	old_execute_ex (execute_data TSRMLS_CC);
+#else
 	if (zo) {
 		old_execute_ZO (op_array, dummy TSRMLS_CC);
 	} else {
 		old_execute (op_array TSRMLS_CC);
 	}
+#endif
 	/* nothing to do */
 	SUHOSIN_G(in_code_type) = orig_code_type;
 	SUHOSIN_G(execution_depth)--;
 }
 /* }}} */
 
+#if PHP_VERSION_ID < 50500
 /* {{{ void suhosin_execute(zend_op_array *op_array TSRMLS_DC)
  *    This function provides a hook for execution */
 static void suhosin_execute(zend_op_array *op_array TSRMLS_DC)
@@ -577,11 +654,17 @@ static void suhosin_execute_ZO(zend_op_array *op_array, long dummy TSRMLS_DC)
 	suhosin_execute_ex(op_array, 1, dummy TSRMLS_CC);
 }	
 /* }}} */
+#endif
 
-
+#if PHP_VERSION_ID >= 50500
+#define IH_HANDLER_PARAMS_REST int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used TSRMLS_DC
+#define IH_HANDLER_PARAMS internal_function_handler *ih, IH_HANDLER_PARAMS_REST
+#define IH_HANDLER_PARAM_PASSTHRU ih, ht, return_value, return_value_ptr, this_ptr, return_value_used TSRMLS_CC
+#else
 #define IH_HANDLER_PARAMS_REST zend_execute_data *execute_data_ptr, int return_value_used, int ht, zval *return_value TSRMLS_DC
 #define IH_HANDLER_PARAMS internal_function_handler *ih, IH_HANDLER_PARAMS_REST
 #define IH_HANDLER_PARAM_PASSTHRU ih, execute_data_ptr, return_value_used, ht, return_value TSRMLS_CC
+#endif
 
 HashTable ihandler_table;
 
@@ -600,11 +683,11 @@ int ih_preg_replace(IH_HANDLER_PARAMS)
 	zval **regex,
 	     **replace,
 	     **subject,
-	     **limit;
+	     **limit, **zcount;
 
-	if (ZEND_NUM_ARGS() < 3 || zend_get_parameters_ex(3, &regex, &replace, &subject, &limit) == FAILURE) {
-		return (0);
-	}
+	 if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ZZZ|ZZ", &regex, &replace, &subject, &limit, &zcount) == FAILURE) {
+	 	return(0);
+	 }
 		
 	if (Z_TYPE_PP(regex) == IS_ARRAY) {
 		zval	**regex_entry;
@@ -763,7 +846,11 @@ int ih_mail(IH_HANDLER_PARAMS)
 
 int ih_querycheck(IH_HANDLER_PARAMS)
 {
+#ifdef PHP_ATLEAST_5_3
+    void **p = zend_vm_stack_top(TSRMLS_C) - 1;
+#else
 	void **p = EG(argument_stack).top_element-2;
+#endif
 	unsigned long arg_count;
 	zval **arg;
 	char *query, *s, *e;
@@ -919,7 +1006,11 @@ int ih_querycheck(IH_HANDLER_PARAMS)
 
 int ih_fixusername(IH_HANDLER_PARAMS)
 {
+#ifdef PHP_ATLEAST_5_3
+    void **p = zend_vm_stack_top(TSRMLS_C) - 1;
+#else
 	void **p = EG(argument_stack).top_element-2;
+#endif
 	unsigned long arg_count;
 	zval **arg;char *prefix, *postfix, *user;
 	zval *backup, *my_user;
@@ -987,56 +1078,465 @@ int ih_fixusername(IH_HANDLER_PARAMS)
 	return (0);
 }
 
-static int suhosin_php_body_write(const char *str, uint str_length TSRMLS_DC)
+
+static int ih_function_exists(IH_HANDLER_PARAMS)
 {
-#define P_META_ROBOTS "<meta name=\"ROBOTS\" content=\"NOINDEX,NOFOLLOW,NOARCHIVE\" />"
-#define S_META_ROBOTS "<meta name=\"ROBOTS\" content=\"NOINDEX,FOLLOW,NOARCHIVE\" />"
-
-    SDEBUG("bw: %s", str);
-
-	if ((str_length == sizeof("</head>\n")-1) && (strcmp(str, "</head>\n")==0)) {
-		SUHOSIN_G(old_php_body_write)(S_META_ROBOTS, sizeof(S_META_ROBOTS)-1 TSRMLS_CC);
-		OG(php_body_write) = SUHOSIN_G(old_php_body_write);
-		return SUHOSIN_G(old_php_body_write)(str, str_length TSRMLS_CC);
-	} else if ((str_length == sizeof(P_META_ROBOTS)-1) && (strcmp(str, P_META_ROBOTS)==0)) {
-		return str_length;
+#ifndef PHP_ATLEAST_5_3
+	zval **function_name;
+#endif
+	zend_function *func;
+	char *lcname;
+	zend_bool retval;
+	int func_name_len;
+	
+#ifndef PHP_ATLEAST_5_3
+	if (ZEND_NUM_ARGS()!=1 || zend_get_parameters_ex(1, &function_name)==FAILURE) {
+		ZEND_WRONG_PARAM_COUNT_WITH_RETVAL(1);
 	}
-	return SUHOSIN_G(old_php_body_write)(str, str_length TSRMLS_CC);	
-}
-
-static int ih_phpinfo(IH_HANDLER_PARAMS)
-{
-    int argc = ZEND_NUM_ARGS();
-	long flag;
-
-	if (zend_parse_parameters(argc TSRMLS_CC, "|l", &flag) == FAILURE) {
-        RETVAL_FALSE;
+	convert_to_string_ex(function_name);
+	func_name_len = Z_STRLEN_PP(function_name);
+	lcname = estrndup(Z_STRVAL_PP(function_name), func_name_len);	
+	zend_str_tolower(lcname, func_name_len);
+#else
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &lcname, &func_name_len) == FAILURE) {
 		return (1);
 	}
 
-	if(!argc) {
-		flag = PHP_INFO_ALL;
+	/* Ignore leading "\" */
+	if (func_name_len > 0 && lcname[0] == '\\') {
+		lcname = &lcname[1];
+		func_name_len--;
+	}
+	lcname = zend_str_tolower_dup(lcname, func_name_len);	
+#endif
+
+	retval = (zend_hash_find(EG(function_table), lcname, func_name_len+1, (void **)&func) == SUCCESS);
+	
+	/*
+	 * A bit of a hack, but not a bad one: we see if the handler of the function
+	 * is actually one that displays "function is disabled" message.
+	 */
+	if (retval && func->type == ZEND_INTERNAL_FUNCTION &&
+		func->internal_function.handler == zif_display_disabled_function) {
+		retval = 0;
 	}
 
-	/* Andale!  Andale!  Yee-Hah! */
-	php_start_ob_buffer(NULL, 4096, 0 TSRMLS_CC);
-	if (!sapi_module.phpinfo_as_text) {
-		SUHOSIN_G(old_php_body_write) = OG(php_body_write);
-		OG(php_body_write) = suhosin_php_body_write;
+	/* Now check if function is forbidden by Suhosin */
+	if (SUHOSIN_G(in_code_type) == SUHOSIN_EVAL) {
+		if (SUHOSIN_G(eval_whitelist) != NULL) {
+			if (!zend_hash_exists(SUHOSIN_G(eval_whitelist), lcname, func_name_len+1)) {
+			    retval = 0;
+			}
+		} else if (SUHOSIN_G(eval_blacklist) != NULL) {
+			if (zend_hash_exists(SUHOSIN_G(eval_blacklist), lcname, func_name_len+1)) {
+			    retval = 0;
+			}
+		}
+	}
+	
+	if (SUHOSIN_G(func_whitelist) != NULL) {
+		if (!zend_hash_exists(SUHOSIN_G(func_whitelist), lcname, func_name_len+1)) {
+		    retval = 0;
+		}
+	} else if (SUHOSIN_G(func_blacklist) != NULL) {
+		if (zend_hash_exists(SUHOSIN_G(func_blacklist), lcname, func_name_len+1)) {
+		    retval = 0;
+		}
 	}
-	php_print_info(flag TSRMLS_CC);
-	php_end_ob_buffer(1, 0 TSRMLS_CC);
 
-	RETVAL_TRUE;
+	efree(lcname);
+
+	RETVAL_BOOL(retval);
 	return (1);
 }
 
+/* MT RAND FUNCTIONS */
+
+/*
+	The following php_mt_...() functions are based on a C++ class MTRand by
+	Richard J. Wagner. For more information see the web page at
+	http://www-personal.engin.umich.edu/~wagnerr/MersenneTwister.html
+
+	Mersenne Twister random number generator -- a C++ class MTRand
+	Based on code by Makoto Matsumoto, Takuji Nishimura, and Shawn Cokus
+	Richard J. Wagner  v1.0  15 May 2003  rjwagner@writeme.com
+
+	The Mersenne Twister is an algorithm for generating random numbers.  It
+	was designed with consideration of the flaws in various other generators.
+	The period, 2^19937-1, and the order of equidistribution, 623 dimensions,
+	are far greater.  The generator is also fast; it avoids multiplication and
+	division, and it benefits from caches and pipelines.  For more information
+	see the inventors' web page at http://www.math.keio.ac.jp/~matumoto/emt.html
+
+	Reference
+	M. Matsumoto and T. Nishimura, "Mersenne Twister: A 623-Dimensionally
+	Equidistributed Uniform Pseudo-Random Number Generator", ACM Transactions on
+	Modeling and Computer Simulation, Vol. 8, No. 1, January 1998, pp 3-30.
+
+	Copyright (C) 1997 - 2002, Makoto Matsumoto and Takuji Nishimura,
+	Copyright (C) 2000 - 2003, Richard J. Wagner
+	All rights reserved.                          
+
+	Redistribution and use in source and binary forms, with or without
+	modification, are permitted provided that the following conditions
+	are met:
+
+	1. Redistributions of source code must retain the above copyright
+	   notice, this list of conditions and the following disclaimer.
+
+	2. Redistributions in binary form must reproduce the above copyright
+	   notice, this list of conditions and the following disclaimer in the
+	   documentation and/or other materials provided with the distribution.
+
+	3. The names of its contributors may not be used to endorse or promote 
+	   products derived from this software without specific prior written 
+	   permission.
+
+	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+	"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+	LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+	A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
+	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+	PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+	NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+
+	The original code included the following notice:
+
+	When you use this, send an email to: matumoto@math.keio.ac.jp
+    with an appropriate reference to your work.
+
+	It would be nice to CC: rjwagner@writeme.com and Cokus@math.washington.edu
+	when you write.
+*/
+
+#define N             624                 /* length of state vector */
+#define M             (397)                /* a period parameter */
+#define hiBit(u)      ((u) & 0x80000000U)  /* mask all but highest   bit of u */
+#define loBit(u)      ((u) & 0x00000001U)  /* mask all but lowest    bit of u */
+#define loBits(u)     ((u) & 0x7FFFFFFFU)  /* mask     the highest   bit of u */
+#define mixBits(u, v) (hiBit(u)|loBits(v)) /* move hi bit of u to hi bit of v */
+
+#define twist(m,u,v)  (m ^ (mixBits(u,v)>>1) ^ ((php_uint32)(-(php_int32)(loBit(u))) & 0x9908b0dfU))
+
+/* {{{ php_mt_initialize
+ */
+static inline void suhosin_mt_initialize(php_uint32 seed, php_uint32 *state)
+{
+	/* Initialize generator state with seed
+	   See Knuth TAOCP Vol 2, 3rd Ed, p.106 for multiplier.
+	   In previous versions, most significant bits (MSBs) of the seed affect
+	   only MSBs of the state array.  Modified 9 Jan 2002 by Makoto Matsumoto. */
+
+	register php_uint32 *s = state;
+	register php_uint32 *r = state;
+	register int i = 1;
+
+	*s++ = seed & 0xffffffffU;
+	for( ; i < N; ++i ) {
+		*s++ = ( 1812433253U * ( *r ^ (*r >> 30) ) + i ) & 0xffffffffU;
+		r++;
+	}
+}
+/* }}} */
+
+static inline void suhosin_mt_init_by_array(php_uint32 *key, int keylen, php_uint32 *state)
+{
+    int i, j, k;
+    suhosin_mt_initialize(19650218U, state);
+    i = 1; j = 0;
+    k = (N > keylen ? N : keylen);
+    for (; k; k--) {
+        state[i] = (state[i] ^ ((state[i-1] ^ (state[i-1] >> 30)) * 1664525U)) + key[j] + j;
+        i++; j = (j+1) % keylen;
+        if (i >= N) { state[0] = state[N-1]; i=1; }
+    }
+    for (k=N-1; k; k--) {
+        state[i] = (state[i] ^ ((state[i-1] ^ (state[i-1] >> 30)) * 1566083941U)) - i;
+        i++;
+        if (i >= N) { state[0] = state[N-1]; i=1; }
+    }
+    state[0] = 0x80000000U;
+}
+/* }}} */
+
+
+/* {{{ suhosin_mt_reload
+ */
+static inline void suhosin_mt_reload(php_uint32 *state, php_uint32 **next, int *left)
+{
+	/* Generate N new values in state
+	   Made clearer and faster by Matthew Bellew (matthew.bellew@home.com) */
+
+	register php_uint32 *p = state;
+	register int i;
+
+	for (i = N - M; i--; ++p)
+		*p = twist(p[M], p[0], p[1]);
+	for (i = M; --i; ++p)
+		*p = twist(p[M-N], p[0], p[1]);
+	*p = twist(p[M-N], p[0], state[0]);
+	*left = N;
+	*next = state;
+}
+/* }}} */
+
+/* {{{ suhosin_mt_srand
+ */
+static void suhosin_mt_srand(php_uint32 seed TSRMLS_DC)
+{
+	/* Seed the generator with a simple uint32 */
+	suhosin_mt_initialize(seed, SUHOSIN_G(mt_state));
+	suhosin_mt_reload(SUHOSIN_G(mt_state), &SUHOSIN_G(mt_next), &SUHOSIN_G(mt_left));
+
+	/* Seed only once */
+	SUHOSIN_G(mt_is_seeded) = 1;
+}
+/* }}} */
+
+/* {{{ suhosin_mt_rand
+ */
+static php_uint32 suhosin_mt_rand(TSRMLS_D)
+{
+	/* Pull a 32-bit integer from the generator state
+	   Every other access function simply transforms the numbers extracted here */
+	
+	register php_uint32 s1;
+
+	if (SUHOSIN_G(mt_left) == 0) {
+    	suhosin_mt_reload(SUHOSIN_G(mt_state), &SUHOSIN_G(mt_next), &SUHOSIN_G(mt_left));
+	}
+	--SUHOSIN_G(mt_left);
+		
+	s1 = *SUHOSIN_G(mt_next)++;
+	s1 ^= (s1 >> 11);
+	s1 ^= (s1 <<  7) & 0x9d2c5680U;
+	s1 ^= (s1 << 15) & 0xefc60000U;
+	return ( s1 ^ (s1 >> 18) );
+}
+/* }}} */
+
+/* {{{ suhosin_gen_entropy
+ */
+static void suhosin_gen_entropy(php_uint32 *seedbuf TSRMLS_DC)
+{
+    /* On a modern OS code, stack and heap base are randomized */
+    unsigned long code_value  = (unsigned long)suhosin_gen_entropy;
+    unsigned long stack_value = (unsigned long)&code_value;
+    unsigned long heap_value  = (unsigned long)SUHOSIN_G(r_state);
+    suhosin_SHA256_CTX   context;
+    int fd;
+    
+    code_value ^= code_value >> 32;
+    stack_value ^= stack_value >> 32;
+    heap_value ^= heap_value >> 32;
+    
+    seedbuf[0] = code_value;
+    seedbuf[1] = stack_value;
+    seedbuf[2] = heap_value;
+    seedbuf[3] = time(0);
+#ifdef PHP_WIN32
+    seedbuf[4] = GetCurrentProcessId();
+#else
+    seedbuf[4] = getpid();
+#endif
+    seedbuf[5] = (php_uint32) 0x7fffffff * php_combined_lcg(TSRMLS_C);
+
+#ifndef PHP_WIN32
+    fd = VCWD_OPEN("/dev/urandom", O_RDONLY);
+    if (fd >= 0) {
+        /* ignore error case - if urandom doesn't give us any/enough random bytes */
+        read(fd, &seedbuf[6], 2 * sizeof(php_uint32));
+        close(fd);
+    }
+#endif
+
+    suhosin_SHA256Init(&context);
+	suhosin_SHA256Update(&context, (void *) seedbuf, sizeof(php_uint32) * 8);
+	suhosin_SHA256Final((void *)seedbuf, &context);
+}
+/* }}} */
+
+
+/* {{{ suhosin_srand_auto
+ */
+static void suhosin_srand_auto(TSRMLS_D)
+{
+    php_uint32 seed[8];    
+    suhosin_gen_entropy(&seed[0] TSRMLS_CC);
+
+	suhosin_mt_init_by_array(seed, 8, SUHOSIN_G(r_state));
+	suhosin_mt_reload(SUHOSIN_G(r_state), &SUHOSIN_G(r_next), &SUHOSIN_G(r_left));
+
+	/* Seed only once */
+	SUHOSIN_G(r_is_seeded) = 1;
+}
+/* }}} */
+
+/* {{{ suhosin_mt_srand_auto
+ */
+static void suhosin_mt_srand_auto(TSRMLS_D)
+{
+    php_uint32 seed[8];    
+    suhosin_gen_entropy(&seed[0] TSRMLS_CC);
+
+	suhosin_mt_init_by_array(seed, 8, SUHOSIN_G(mt_state));
+	suhosin_mt_reload(SUHOSIN_G(mt_state), &SUHOSIN_G(mt_next), &SUHOSIN_G(mt_left));
+
+	/* Seed only once */
+	SUHOSIN_G(mt_is_seeded) = 1;
+}
+/* }}} */
+
+
+/* {{{ suhosin_srand
+ */
+static void suhosin_srand(php_uint32 seed TSRMLS_DC)
+{
+	/* Seed the generator with a simple uint32 */
+	suhosin_mt_initialize(seed+0x12345, SUHOSIN_G(r_state));
+	suhosin_mt_reload(SUHOSIN_G(r_state), &SUHOSIN_G(r_next), &SUHOSIN_G(r_left));
+
+	/* Seed only once */
+	SUHOSIN_G(r_is_seeded) = 1;
+}
+/* }}} */
+
+/* {{{ suhosin_mt_rand
+ */
+static php_uint32 suhosin_rand(TSRMLS_D)
+{
+	/* Pull a 32-bit integer from the generator state
+	   Every other access function simply transforms the numbers extracted here */
+	
+	register php_uint32 s1;
+
+	if (SUHOSIN_G(r_left) == 0) {
+    	suhosin_mt_reload(SUHOSIN_G(r_state), &SUHOSIN_G(r_next), &SUHOSIN_G(r_left));
+	}
+	--SUHOSIN_G(r_left);
+		
+	s1 = *SUHOSIN_G(r_next)++;
+	s1 ^= (s1 >> 11);
+	s1 ^= (s1 <<  7) & 0x9d2c5680U;
+	s1 ^= (s1 << 15) & 0xefc60000U;
+	return ( s1 ^ (s1 >> 18) );
+}
+/* }}} */
+
+static int ih_srand(IH_HANDLER_PARAMS)
+{
+    int argc = ZEND_NUM_ARGS();
+	long seed;
+
+	if (zend_parse_parameters(argc TSRMLS_CC, "|l", &seed) == FAILURE || SUHOSIN_G(srand_ignore)) {
+    	return (1);
+    }
+
+    if (argc == 0) {
+        suhosin_srand_auto(TSRMLS_C);
+    } else {
+        suhosin_srand(seed TSRMLS_CC);
+    }
+	return (1);
+}
+
+static int ih_mt_srand(IH_HANDLER_PARAMS)
+{
+    int argc = ZEND_NUM_ARGS();
+	long seed;
+
+	if (zend_parse_parameters(argc TSRMLS_CC, "|l", &seed) == FAILURE || SUHOSIN_G(mt_srand_ignore)) {
+    	return (1);
+    }
+    
+    if (argc == 0) {
+        suhosin_mt_srand_auto(TSRMLS_C);
+    } else {
+        suhosin_mt_srand(seed TSRMLS_CC);
+    }
+	return (1);
+}
+
+static int ih_mt_rand(IH_HANDLER_PARAMS)
+{
+    int argc = ZEND_NUM_ARGS();
+    long min;
+	long max;
+	long number;
+
+	if (argc != 0 && zend_parse_parameters(argc TSRMLS_CC, "ll", &min, &max) == FAILURE) {
+	    return (1);        
+	}
+
+	if (!SUHOSIN_G(mt_is_seeded)) {
+		suhosin_mt_srand_auto(TSRMLS_C);
+	}
+
+	number = (long) (suhosin_mt_rand(TSRMLS_C) >> 1);
+	if (argc == 2) {
+		RAND_RANGE(number, min, max, PHP_MT_RAND_MAX);
+	}
+
+	RETVAL_LONG(number);
+        return (1);
+}
+
+static int ih_rand(IH_HANDLER_PARAMS)
+{
+    int argc = ZEND_NUM_ARGS();
+    long min;
+	long max;
+	long number;
+
+	if (argc != 0 && zend_parse_parameters(argc TSRMLS_CC, "ll", &min, &max) == FAILURE) {
+	    return (1);        
+	}
+
+	if (!SUHOSIN_G(r_is_seeded)) {
+		suhosin_srand_auto(TSRMLS_C);
+	}
+
+	number = (long) (suhosin_rand(TSRMLS_C) >> 1);
+	if (argc == 2) {
+		RAND_RANGE(number, min, max, PHP_MT_RAND_MAX);
+	}
+
+	RETVAL_LONG(number);
+        return (1);
+}
+
+static int ih_getrandmax(IH_HANDLER_PARAMS)
+{
+#ifdef PHP_ATLEAST_5_3
+	if (zend_parse_parameters_none() == FAILURE) {
+		return(0);
+	}
+#else
+        int argc = ZEND_NUM_ARGS();
+
+        if (argc != 0) {
+		ZEND_WRONG_PARAM_COUNT_WITH_RETVAL(1);
+        }
+#endif    
+	RETVAL_LONG(PHP_MT_RAND_MAX);
+	return (1);
+}
 
 internal_function_handler ihandlers[] = {
     { "preg_replace", ih_preg_replace, NULL, NULL, NULL },
     { "mail", ih_mail, NULL, NULL, NULL },
     { "symlink", ih_symlink, NULL, NULL, NULL },
-    { "phpinfo", ih_phpinfo, NULL, NULL, NULL },
+	
+	{ "srand", ih_srand, NULL, NULL, NULL },
+	{ "mt_srand", ih_mt_srand, NULL, NULL, NULL },
+	{ "rand", ih_rand, NULL, NULL, NULL },
+	{ "mt_rand", ih_mt_rand, NULL, NULL, NULL },
+	{ "getrandmax", ih_getrandmax, NULL, NULL, NULL },
+	{ "mt_getrandmax", ih_getrandmax, NULL, NULL, NULL },
 	
     { "ocilogon", ih_fixusername, (void *)1, NULL, NULL },
     { "ociplogon", ih_fixusername, (void *)1, NULL, NULL },
@@ -1048,6 +1548,8 @@ internal_function_handler ihandlers[] = {
     { "fbsql_change_user", ih_fixusername, (void *)1, NULL, NULL },
     { "fbsql_connect", ih_fixusername, (void *)2, NULL, NULL },
     { "fbsql_pconnect", ih_fixusername, (void *)2, NULL, NULL },
+    
+    { "function_exists", ih_function_exists, NULL, NULL, NULL },
 	
     { "ifx_connect", ih_fixusername, (void *)2, NULL, NULL },
     { "ifx_pconnect", ih_fixusername, (void *)2, NULL, NULL },
@@ -1082,25 +1584,74 @@ internal_function_handler ihandlers[] = {
 };
 
 #define FUNCTION_WARNING() zend_error(E_WARNING, "%s() has been disabled for security reasons", get_active_function_name(TSRMLS_C));
+#define FUNCTION_SIMULATE_WARNING() zend_error(E_WARNING, "SIMULATION - %s() has been disabled for security reasons", get_active_function_name(TSRMLS_C));
 
 /* {{{ void suhosin_execute_internal(zend_execute_data *execute_data_ptr, int return_value_used TSRMLS_DC)
  *    This function provides a hook for internal execution */
+#if PHP_VERSION_ID >= 50500
+#define EX_T(offset) (*EX_TMP_VAR(execute_data_ptr, offset))
+
+static void suhosin_execute_internal(zend_execute_data *execute_data_ptr, zend_fcall_info *fci, int return_value_used TSRMLS_DC)
+{
+	zval *return_value;
+	zval **return_value_ptr;
+	zval *this_ptr;
+	int ht;
+	
+	if (fci) {
+		return_value = *fci->retval_ptr_ptr;
+		return_value_ptr = fci->retval_ptr_ptr;
+		this_ptr = fci->object_ptr;
+		ht = fci->param_count;
+	} else {
+		temp_variable *ret = &EX_T(execute_data_ptr->opline->result.var);
+		zend_function *fbc = execute_data_ptr->function_state.function;
+		return_value = ret->var.ptr;
+		return_value_ptr = (fbc->common.fn_flags & ZEND_ACC_RETURN_REFERENCE) ? &ret->var.ptr : NULL;
+		this_ptr = execute_data_ptr->object;
+		ht = execute_data_ptr->opline->extended_value;
+	}	
+#else
 static void suhosin_execute_internal(zend_execute_data *execute_data_ptr, int return_value_used TSRMLS_DC)
 {
-	char *lcname;
-	int function_name_strlen;
 	zval *return_value;
-	int ht;
+	int ht = execute_data_ptr->opline->extended_value;
+#endif
+	char *lcname;
+	int function_name_strlen, free_lcname = 0;
+	zend_class_entry *ce = NULL;
 	internal_function_handler *ih;
+
+#ifdef ZEND_ENGINE_2
+	ce = ((zend_internal_function *) execute_data_ptr->function_state.function)->scope;
+#endif
+	lcname = (char *)((zend_internal_function *) execute_data_ptr->function_state.function)->function_name;
+	function_name_strlen = strlen(lcname);
 	
-	lcname = ((zend_internal_function *) execute_data_ptr->function_state.function)->function_name;
-	function_name_strlen = strlen(lcname);
+	/* handle methodcalls correctly */
+	if (ce != NULL) {
+		char *tmp = (char *) emalloc(function_name_strlen + 2 + ce->name_length + 1);
+		memcpy(tmp, ce->name, ce->name_length);
+		memcpy(tmp+ce->name_length, "::", 2);
+		memcpy(tmp+ce->name_length+2, lcname, function_name_strlen);
+		lcname = tmp;
+		free_lcname = 1;
+		function_name_strlen += ce->name_length + 2;
+		lcname[function_name_strlen] = 0;
+		zend_str_tolower(lcname, function_name_strlen);
+	}
+
+#if PHP_VERSION_ID < 50500	
 #ifdef ZEND_ENGINE_2  
+# if PHP_VERSION_ID < 50400
 	return_value = (*(temp_variable *)((char *) execute_data_ptr->Ts + execute_data_ptr->opline->result.u.var)).var.ptr;
+# else
+	return_value = (*(temp_variable *)((char *) execute_data_ptr->Ts + execute_data_ptr->opline->result.var)).var.ptr;
+# endif
 #else
-    return_value = execute_data_ptr->Ts[execute_data_ptr->opline->result.u.var].var.ptr;
+        return_value = execute_data_ptr->Ts[execute_data_ptr->opline->result.u.var].var.ptr;
+#endif
 #endif
-	ht = execute_data_ptr->opline->extended_value;
 
 	SDEBUG("function: %s", lcname);
 
@@ -1108,30 +1659,42 @@ static void suhosin_execute_internal(zend_execute_data *execute_data_ptr, int re
 	
 		if (SUHOSIN_G(eval_whitelist) != NULL) {
 			if (!zend_hash_exists(SUHOSIN_G(eval_whitelist), lcname, function_name_strlen+1)) {
-			    suhosin_log(S_EXECUTOR, "function outside of eval whitelist called: %s()", lcname);
-			    FUNCTION_WARNING()
-			    suhosin_bailout(TSRMLS_C);
+				suhosin_log(S_EXECUTOR|S_GETCALLER, "function outside of eval whitelist called: %s()", lcname);
+				if (!SUHOSIN_G(simulation)) {
+				        goto execute_internal_bailout;
+        			} else {
+        			        FUNCTION_SIMULATE_WARNING()
+				}
 			}
 		} else if (SUHOSIN_G(eval_blacklist) != NULL) {
 			if (zend_hash_exists(SUHOSIN_G(eval_blacklist), lcname, function_name_strlen+1)) {
-			    suhosin_log(S_EXECUTOR, "function within eval blacklist called: %s()", lcname);
-			    FUNCTION_WARNING()
-			    suhosin_bailout(TSRMLS_C);
+				suhosin_log(S_EXECUTOR|S_GETCALLER, "function within eval blacklist called: %s()", lcname);
+				if (!SUHOSIN_G(simulation)) {
+				        goto execute_internal_bailout;
+        			} else {
+        			        FUNCTION_SIMULATE_WARNING()
+				}
 			}
 		}
 	}
 	
 	if (SUHOSIN_G(func_whitelist) != NULL) {
 		if (!zend_hash_exists(SUHOSIN_G(func_whitelist), lcname, function_name_strlen+1)) {
-		    suhosin_log(S_EXECUTOR, "function outside of whitelist called: %s()", lcname);
-		    FUNCTION_WARNING()
-		    suhosin_bailout(TSRMLS_C);
+			suhosin_log(S_EXECUTOR|S_GETCALLER, "function outside of whitelist called: %s()", lcname);
+			if (!SUHOSIN_G(simulation)) {
+			        goto execute_internal_bailout;
+			} else {
+			        FUNCTION_SIMULATE_WARNING()
+			}
 		}
 	} else if (SUHOSIN_G(func_blacklist) != NULL) {
 		if (zend_hash_exists(SUHOSIN_G(func_blacklist), lcname, function_name_strlen+1)) {
-		    suhosin_log(S_EXECUTOR, "function within blacklist called: %s()", lcname);
-		    FUNCTION_WARNING()
-		    suhosin_bailout(TSRMLS_C);
+			suhosin_log(S_EXECUTOR|S_GETCALLER, "function within blacklist called: %s()", lcname);
+			if (!SUHOSIN_G(simulation)) {
+			        goto execute_internal_bailout;
+			} else {
+			        FUNCTION_SIMULATE_WARNING()
+			}
 		}
 	}
 	
@@ -1145,11 +1708,29 @@ static void suhosin_execute_internal(zend_execute_data *execute_data_ptr, int re
 		}
 		
 		if (retval == 0) {
+#if PHP_VERSION_ID >= 50500
+			old_execute_internal(execute_data_ptr, fci, return_value_used TSRMLS_CC);
+#else
 			old_execute_internal(execute_data_ptr, return_value_used TSRMLS_CC);
+#endif
 		}
 	} else {
+#if PHP_VERSION_ID >= 50500
+		old_execute_internal(execute_data_ptr, fci, return_value_used TSRMLS_CC);
+#else
 		old_execute_internal(execute_data_ptr, return_value_used TSRMLS_CC);
+#endif
 	}
+	if (free_lcname == 1) {
+		efree(lcname);
+	}
+	return;
+execute_internal_bailout:
+	if (free_lcname == 1) {
+		efree(lcname);
+	}
+	FUNCTION_WARNING()
+	suhosin_bailout(TSRMLS_C);
 }
 /* }}} */
 
@@ -1178,13 +1759,19 @@ static int function_lookup(zend_extension *extension)
 void suhosin_hook_execute(TSRMLS_D)
 {
 	internal_function_handler *ih;
-	
+
+#if PHP_VERSION_ID >= 50500
+	old_execute_ex = zend_execute_ex;
+	zend_execute_ex = suhosin_execute_ex;
+#else	
 	old_execute = zend_execute;
 	zend_execute = suhosin_execute;
+#endif
 	
 /*	old_compile_file = zend_compile_file;
 	zend_compile_file = suhosin_compile_file; */
 
+#if ZO_COMPATIBILITY_HACK_TEMPORARY_DISABLED
 	if (zo_set_oe_ex == NULL) {	
 		zo_set_oe_ex = (void *)DL_FETCH_SYMBOL(NULL, "zend_optimizer_set_oe_ex");
 	}
@@ -1195,6 +1782,7 @@ void suhosin_hook_execute(TSRMLS_D)
 	if (zo_set_oe_ex != NULL) {
 		old_execute_ZO = zo_set_oe_ex(suhosin_execute_ZO);
 	}
+#endif
 	
 	old_execute_internal = zend_execute_internal;
 	if (old_execute_internal == NULL) {
@@ -1229,12 +1817,18 @@ void suhosin_hook_execute(TSRMLS_D)
  */
 void suhosin_unhook_execute()
 {
+#if ZO_COMPATIBILITY_HACK_TEMPORARY_DISABLED
 	if (zo_set_oe_ex) {
 		zo_set_oe_ex(old_execute_ZO);
 	}
-	
+#endif
+
+#if PHP_VERSION_ID >= 50500	
+	zend_execute_ex = old_execute_ex;
+#else
 	zend_execute = old_execute;
-	
+#endif
+		
 /*	zend_compile_file = old_compile_file; */
 
 	if (old_execute_internal == execute_internal) {
